import pydantic
from pydantic import BaseModel
from typing import List, Union, Dict, Any, Optional
from enum import Enum
import logging
from bson.objectid import ObjectId
from json import JSONEncoder
import json

def _default(self, obj):
    return getattr(obj.__class__, "toJSON", _default.default)(obj)

_default.default = JSONEncoder().default
JSONEncoder.default = _default

class Id(ObjectId):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not isinstance(v, ObjectId) and not isinstance(v, str):
            raise TypeError('ObjectId required')
        try:
            logging.info('validating..')
            res = Id(v)
            logging.info(type(res))
            return res
        except bson.errors.InvalidId:
            logging.warning(f'{v} is not ObjectId')
            raise TypeError(f'{v} is not ObjectId')


    @classmethod
    def __modify_schema__(cls, schema):
        schema.update({
            'Title': 'MongoDB ObjectID',
            'type': 'string'
        })

    def toJSON(self):
        return str(self)

class EntityType(Enum):
    """https://cloud.google.com/natural-language/docs/reference/rest/v1beta2/Entity#Type"""
    UNKNOWN = "UNKNOWN"
    PERSON = "PERSON"
    LOCATION = "LOCATION"
    ORGANIZATION = "ORGANIZATION"
    EVENT = "EVENT"
    WORK_OF_ART = "WORK_OF_ART"
    CONSUMER_GOOD = "CONSUMER_GOOD"
    OTHER = "OTHER"
    PHONE_NUMBER = "PHONE_NUMBER"
    ADDRESS = "ADDRESS"
    DATE = "DATE"
    NUMBER = "NUMBER"
    PRICE = "PRICE"

class Sentiment(BaseModel):
    score: float
    magnitude: float

class Entity(BaseModel):
    name: str
    type: EntityType
    meta: Dict[str, str]
    is_named: bool
    sentiment: Sentiment
    salience: float

class Token(BaseModel):
    raw: str
    offset: int # position in the original phrase
    pos: Dict[Any, Any] # Part of Speech; Make Enum

class Language(str, Enum):
    en = "en"
    other = "other"

class Category(BaseModel):
    """https://cloud.google.com/natural-language/docs/categories"""
    name: str
    confidence: float

class Speech(BaseModel):
    raw: str
    tokenized: List[Token]
    stylized: str
    args: List[Entity]
    sentiment: Union[Sentiment, None]
    categories: List[Category]
    lang: Language = Language.en

class Author(BaseModel):
    is_user: bool
    _id: str = ""

    class Config:
        extra = pydantic.Extra.allow

class Replic(BaseModel):
    user: Author
    author: Author
    timestamp: int
    speech: Speech
    _id: Optional[Id] = None

    class Config:
        extra = pydantic.Extra.allow

class SessionConfigs(BaseModel):
    local: bool = True
    visual: bool = False

class UserSettings(BaseModel):
    style: bool = False
    confirmation_threshold: float = 1.0

class UserState(BaseModel):
    dialog_skill_id: str = ""
    is_waiting: bool = False

class PublicUser(BaseModel):
    session: SessionConfigs = SessionConfigs()
    settings: UserSettings = UserSettings()
    state: UserState = UserState()

class User(PublicUser):
    token: Optional[str] = None
    _id: Optional[Id] = None
    telegram_id: Optional[int] = None

    class Config:
        extra = pydantic.Extra.allow

class Skill(BaseModel):
    url: str # b64
    author: Author
    name: str
    _id: Optional[Id] = None

    class Config:
        extra = pydantic.Extra.allow

class SkillInstance(BaseModel):
    skill: Skill
    local_state: str # b64
    global_state: str # b64
    user_id: Id
    _id: Optional[Id] = None

    class Config:
        extra = pydantic.Extra.allow



import pymongo
from .types import *

class SkillNotFound(Exception):
    def __init__(self, skill_id):
        self.skill_id = skill_id

    def __str__(self):
        return f"Skill {self.skill_id} not found"

class SkillExists(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return f"Skill {self.name} already exists."

class MongoDB:
    def __init__(self, url: str):
        self.client = pymongo.MongoClient(url)
        self.kadia_db = self.client.kadia_db
        self.users = self.kadia_db.users
        self.history = self.kadia_db.history
        self.skills = self.kadia_db.skills
        self.instances = self.kadia_db.instances

    def fetch_telegram_user(self, telegram_id: int):
        user = self.users.find_one({"telegram_id": telegram_id})
        if user is None:
            user = User(telegram_id=telegram_id)
            user._id = Id(self.users.insert_one(user.dict()).inserted_id)
        else:
            user = User(**user)
        return user

    def add_replic_to_history(self, replic: Replic):
        self.history.insert_one(replic.dict())

    def change_user_waiting_status(self, user: User, status: bool):
        self.users.update_one({'_id': user._id}, {
            '$set': {
                'state.is_waiting': status
            }
        })

    def fetch_skill(self, skill_id: Id):
        skill = self.skills.find_one({'_id': skill_id})
        if skill is None:
            raise SkillNotFound(skill_id)
        return Skill(**skill)

    def fetch_skill_instance(self, skill: Skill, user: User):
        result = self.instances.find_one({'skill._id': skill._id, 'user_id': user._id})
        if result is not None:
            return SkillInstance(result)
        instance = SkillInstance(skill=skill,
                                 local_state=str,
                                 global_state=str,
                                 user_id=User._id)
        instance._id = Id(self.instances.insert_one(instance.dict()).inserted_id)
        return instance

    def add_skill(self, skill: Skill):
        result = self.skills.find_one({"name": skill.name})
        if result is not None:
            raise SkillExists(skill.name)
        return Id(self.skills.insert_one(skill.dict()).inserted_id)

    def _remove_skill(self, name: str):
        self.skills.delete_one({'name': name})

    def update_instance(instance: SkillInstance, local_state: str, global_state: str):
        self.instances.update_one({'_id': instance._id}, {'$set': {'local_state': local_state,
                                                                   'global_state': global_state}})
